import os, sys
from Bio.PDB import PDBParser
from sbiRandM.sbiRandM import *
from sbiRandM.sbiRandM.data import aminoacids, Alphabet

def pdb_to_fasta(pdb_file, out_fasta_file):
    """
    This function takes as input a PDB file, and
    returns the Fasta file of the sequence.
    """

    parser = PDBParser(PERMISSIVE=1)
    structure = parser.get_structure('Complex', pdb_file)

    index = 0
    with open(out_fasta_file, "w") as out_fasta:
        for chain in structure.get_chains():
            out_fasta.write(">" + str(Alphabet[index]) + "\n")
            out_fasta.write(Get_fasta(chain) + "\n")
            index += 1
    
    return out_fasta_file





import sys
sys.path.insert(0,"/usr/local/Cellar/modeller/9.23/modlib")
from modeller import *
from modeller.automodel import *
import glob, os, argparse, warnings
from Bio.PDB.PDBParser import PDBParser
from Bio import SeqIO
from Bio import pairwise2
import shutil

from sbiRandM.sbiRandM.data import modeller_aminoacids as aminoacids
from sbiRandM.sbiRandM.models import Protein_Interaction, Chain, Query
import difflib

warnings.filterwarnings("ignore")


def fasta_to_object(fasta):

    """
    This function takes a fasta file with several chains, and parse it into a
    query object. Also changes the DNA chains to Placeholder characters.

    @input folder - File with Fasta
    @output - Query object.
    """

    record_dict = SeqIO.to_dict(SeqIO.parse(fasta, "fasta"))
    #print (record_dict)
    #print ("Len of record dict", len(record_dict))
    query = Query(name=(os.path.splitext(os.path.basename(fasta))[0]))
    alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    index = 0
    for sequence in record_dict.keys():
        #print (record_dict[sequence], len(record_dict[sequence]))
        chain = Chain(name = alphabet[index], sequence = record_dict[sequence].seq, first_aminoacid = 1, last_aminoacid = len(record_dict[sequence].seq))
        chain.dna_to_placeholder()
        query.add_chain(chain)
        index += 1
    #print ("Fasta to object query:", query.chains)
    return query

def create_models(folder):

    """
    This function takes a folder that contains the complexes of the pairwise interactions, 
    parse them and return a list of Protein_Interaction objects with information 
    about the location and the sequences.

    @input folder - Folder where you store the pairwise complexes of the protein
    @output - List of Protein_Interaction objects
    """

    list_of_interactions = list()
    parser = PDBParser(PERMISSIVE=1)
    set_residues = set()

    for pdb_file in glob.glob(os.path.join(folder,"*.pdb")):
        pdb_name = os.path.basename(pdb_file)
        structure = parser.get_structure('Complex', pdb_file)
        protein = Protein_Interaction(name = pdb_name, path = pdb_file)


        for pdb_chain in structure[0]:
            sequence = ""
            for residue in pdb_chain:
                set_residues.add(residue.get_resname())
                sequence += aminoacids[residue.get_resname().strip()]
            protein.add_chain(Chain(name = pdb_chain.get_id(), sequence = sequence, first_aminoacid = list(pdb_chain)[0].get_id()[1], last_aminoacid = list(pdb_chain)[-1].get_id()[1]))

        list_of_interactions.append(protein)
        #print (protein.name)
    #print (set_residues)
    #for element in set_residues:
    #    print (element, aminoacids[element.strip()])

    #print (list_of_interactions)
    return list_of_interactions

def check_similarity(query, interactions_list):

    """
    This function takes a query object and a list of Pairwise interaction objects,
    Returns the interaction list with the attribute "originalChain" updated for each chain of
    the pairwise interactions.  
    First it tries to check if the sequence of interaction is a subset. If not, it checks over a 73% of similarity
    """
    print (len(query.chains))
    for chain in query.chains:
        #print (chain.name)
        for protein in interactions_list:

            for protein_chain in protein.chains:

                if protein_chain.sequence in chain.sequence:
                    if any(c in chain.sequence for c in "ZJ_OB"):
                        if len(chain.sequence) > len(protein_chain.sequence):
                            chain.sequence = protein_chain.sequence  
                    protein_chain.originalChain.append(chain.name)

                elif difflib.SequenceMatcher(None,chain.sequence.strip(), str(protein_chain.sequence)).ratio() >0.50 :
                    if any(c in chain.sequence for c in "ZJ_OB"):
                        if len(chain.sequence) > len(protein_chain.sequence):
                            chain.sequence = protein_chain.sequence 
                    protein_chain.originalChain.append(chain.name)
                


    return query, interactions_list

def generate_alignment(query, interactions, output_folder):

    if not os.path.isdir(output_folder): os.mkdir(output_folder)

    with open(os.path.join(output_folder, "alignment.pir"), "w") as output:

        # FIRST THE INTERACTIONS
        for interaction in interactions:
            output.write(">P1;" + interaction.name + "\n")
            output.write("structureX:" + 
                        interaction.name + ":" + 
                        interaction.chains[0].first_aminoacid + ":" + 
                        interaction.first_chain + ":" + 
                        interaction.chains[1].last_aminoacid + 
                        ":" + interaction.last_chain + ": : :-1.0:-1.0\n")
            
            for chain in query.chains:
                
                if str(interaction.chains[0].originalChain[0]) == str(chain.name): # CHAIN A OF INTERACTION

                    alignments = pairwise2.align.globalms(chain.sequence, interaction.chains[0].sequence,  2, -1, -30, -10)
                    sequence = alignments[0][1]
                    for letter in "ZB_OJ":
                        sequence = sequence.replace(letter,".")
                    output.write(sequence + "\n/\n")

                elif str(interaction.chains[1].originalChain[0]) == str(chain.name): # CHAIN B OF INTERACTION

                    alignments = pairwise2.align.globalms(chain.sequence, interaction.chains[1].sequence,  2, -1, -30, -10)
                    sequence = alignments[0][1]
                    for letter in "ZB_OJ":
                        sequence = sequence.replace(letter,".")
                    if chain.name == query.chains[-1].name:  
                        output.write(sequence + "*\n\n")
                    else:                    
                        output.write(sequence + "\n/\n")
                
                else:
                    if chain.name == query.chains[-1].name: 
                        output.write("-" * len(chain.sequence) + "*\n\n")
                    else:
                        output.write("-" * len(chain.sequence) + "\n/\n")


        #QUERY
        output.write(">P1;" + query.name.strip() + "\n")
        output.write("sequence:" + query.name + 
                     ": 1:" + query.chains[0].name + 
                     " :" + str(len(query.chains[0].sequence)) + 
                     ":" + query.chains[-1].name + 
                     ": : :-1.0:-1.0\n")
        
        for query_chain in query.chains:
            for letter in "ZB_OJ":
                query_chain.sequence = query_chain.sequence.replace(letter,".")

            if query_chain.name == query.chains[-1].name:  
                output.write(query_chain.sequence.strip() + "*\n\n") 
            else:                    
                output.write(query_chain.sequence.strip() + "\n/\n") 
            
def make_model(output_folder, interaction_pdb_folder, fasta):
    #detect pir file
    templates = list()
    for interaction_file in os.listdir(interaction_pdb_folder):
        if "SEPARED" in os.path.basename(interaction_file):
            templates.append(interaction_file)
            shutil.copyfile( os.path.join(interaction_pdb_folder, interaction_file) , os.path.join(output_folder, interaction_file) )

    os.chdir(output_folder)
    log.verbose()    # request verbose output
    env = environ()  # create a new MODELLER environment to build this model in

    # directories for input atom files
    env.io.atom_files_directory = ['.', '../atom_files']
    pir_file = glob.glob("*.pir")[0]
    env.io.hetatm=True

    a = automodel(env,
                alnfile  = pir_file, # alignment filename
                knowns   = tuple(templates),     # codes of the templates
                sequence = str(os.path.splitext(os.path.basename(fasta))[0].strip()))       

    a.starting_model= 1                 # index of the first model
    a.ending_model  = 1                 # index of the last model

    a.make()                            # do the actual homology modeling

def separe_interactions(interactions):
    """
    This function takes a list of protein pairwise interactions, that have more than originalChain
    (When mapping to homodimers) and divide them into several objects, with one chain for each.
    """
    updated_interactions = list()

    for interaction in interactions:
        
        for original_chain_A in interaction.chains[0].originalChain:
            #print ("This complex enters the first for, original chain", interaction.chains[0].originalChain )
            for original_chain_B in interaction.chains[1].originalChain:
                print (interaction.name, original_chain_A, original_chain_B )
                #print ("This complex enters the second for, original chains", interaction.chains[1].originalChain )

                if original_chain_A == original_chain_B:
                    continue


                Updated_Chain_A = Chain(name = interaction.chains[0].name , 
                                        sequence = interaction.chains[0].sequence, 
                                        first_aminoacid = interaction.chains[0].first_aminoacid,
                                        last_aminoacid = interaction.chains[0].last_aminoacid)
                Updated_Chain_A.originalChain.append(original_chain_A)

                Updated_Chain_B = Chain(name = interaction.chains[1].name , 
                                        sequence = interaction.chains[1].sequence, 
                                        first_aminoacid = interaction.chains[1].first_aminoacid,
                                        last_aminoacid = interaction.chains[1].last_aminoacid)
                Updated_Chain_B.originalChain.append(original_chain_B)


                #COPY THE RESIDUE

                new_interaction_name = "SEPARED_" + interaction.name + "_" + original_chain_A + "_" + original_chain_B + ".pdb"
                new_path =  os.path.join(os.path.dirname(interaction.path), new_interaction_name)
                shutil.copyfile(interaction.path, new_path)

                Updated_interaction = Protein_Interaction(name = new_interaction_name,
                                                          path = new_path)

                Updated_interaction.add_chain(Updated_Chain_A)
                Updated_interaction.add_chain(Updated_Chain_B)

                if Updated_Chain_A.originalChain > Updated_Chain_B.originalChain:
                    Updated_interaction.reversed = True

                updated_interactions.append(Updated_interaction)

    return updated_interactions

def remove_duplicate_chains(folder):
    num_files = 0
    num_deleted_interactions = 0
    actual_list = list()
    for interaction_file in os.listdir(folder):
        num_files += 1
        if "SEPARED" in os.path.basename(interaction_file):
            if (interaction_file[-7], interaction_file[-5]) in actual_list or (interaction_file[-5], interaction_file[-7]) in actual_list:
                os.remove(os.path.join(folder,interaction_file))
                #print ("Deleted interaction:", interaction_file)
                num_deleted_interactions += 1
            else:
                actual_list.append((interaction_file[-7], interaction_file[-5])) 
    print ("Removed", num_deleted_interactions, "files from", num_files, ".")



def reorder_pdb(interactions):
    """
    This function takes a list of object interactions with the attribute Reversed, 
    and reorder the PDB files that has that attribute in True
    """

    for element in interactions:
        if element.reversed == True:
            element.reverse()
            element.chains = list(reversed(element.chains))

            corrected_PDB = os.path.join(os.path.dirname(element.path), "CORRECTED_" + element.name)

            with open(corrected_PDB, "w") as pdb_output:
                with open(element.path, "r") as pdb_input:
                    chain_A = list()

                    for line in pdb_input.readlines():
                        if "ATOM" in line and line[21] == "A":
                            chain_A.append(line)
                        elif "ATOM" in line and line[21] == "B":
                            pdb_output.write(line)
                    
                    pdb_output.write("TER\n")
                    
                    for line in chain_A:
                        pdb_output.write(line)

            os.remove(element.path)
            element.path = corrected_PDB
            element.name = os.path.basename(corrected_PDB)
            element.last_aminoacid = chain_A[-1].split(" ")[5]

            print ("Reversed PDB at path:", corrected_PDB)

def clean_directories(output_folder, TMP_folder):

    to_remove = glob.glob(os.path.join(TMP_folder, "*SEPARED*"))
    to_remove = to_remove + glob.glob(os.path.join(output_folder, "*SEPARED*")) + \
                            glob.glob(os.path.join(output_folder, "*.V999*")) + \
                            glob.glob(os.path.join(output_folder, "*.ini*")) + \
                            glob.glob(os.path.join(output_folder, "*.rsr*")) + \
                            glob.glob(os.path.join(output_folder, "*.D000*")) + \
                            glob.glob(os.path.join(output_folder, "*.sch*")) 

    for file in to_remove:
        os.remove(file)


def mainMod(args):

    query = fasta_to_object(args['fasta_seq'])
    interactions = create_models(args['folder'])
    query, interactions = check_similarity(query, interactions)
    updated_interactions = separe_interactions(interactions)
    reorder_pdb(updated_interactions)
    remove_duplicate_chains(args['folder'])
    
    generate_alignment(query, updated_interactions, args['output_folder'])
    
    make_model(args['output_folder'], args['folder'], args['fasta_seq'])

    #for interaction in interactions:
    #    interaction.pretty_print()

    clean_directories(args['output_folder'], args['folder'])


import os, sys, warnings
import difflib, glob
from Bio.PDB import PDBParser
from sbiRandM.sbiRandM import *
from sbiRandM.sbiRandM.data import aminoacids
from collections import defaultdict
from sbiRandM.sbiRandM.exceptions import *

def check_homology(fasta_1, fasta_2):
    """
    This function check the identity percentage between two chains, and returns True if they are homologous
    or False if they are not. The threshold is currently at 70%.

    @ Input - Fasta_1 : String of fasta sequence
              Fasta_2 : String of fasta sequence

    @ Output - Boolean variable indicating if sequence are homologous
    """

    if fasta_1 in fasta_2:
        return True
    elif fasta_2 in fasta_1:
        return True
    else:
        if difflib.SequenceMatcher(None, fasta_1, fasta_2).ratio() > 0.60:
            return True
        else:
            return False

def Get_fasta(chain):
    """
    This function takes as input a Biopython PDB Chain object, and parse it into 
    a sequence of aminoacids.

    @ Input - Biopython PDB Chain object
    @ Output - String of the sequence of Aminoacids of the PDB Chain Object.
    """

    sequence = ""
    for residue in chain:
        try:
            sequence += aminoacids[residue.get_resname().strip()]
        except:
            continue
    return sequence

def check_clash(structure_complex, mobile_chain):
    """
    This function checks if the new added chain clash with any chain of the complex.

    @ Input - Structure_complex : Biopython object of the actual complex
              Mobile_chain : Biopython object of the moving chain when adding a new chain.

    @ Output - True : Chain is clashing with complex.
               False: Chain is not clashing with complex. 
    """
    backbone_atoms = ["CB", "P"]
    for atom in list(structure_complex.get_atoms()):
        for atom_2 in list(mobile_chain.get_atoms()):
            if atom.id in backbone_atoms and atom_2.id in backbone_atoms:
                if atom_2 - atom < 1:
                    #print ("Ups there is a crash!")
                    return False
    return True

def obtain_pairwise_dict(steichiometry_dict, TMP_folder):

    """
    This function parse a folder full with PDB of the pairwise interactions, and 
    parse them into a dictionary that contains as keys the interacting chains, and
    as value, the path of the file that contains the structure of the interaction.

    @Input - Steichiometry dict with the next format.
              Name of the chain : 
                        Steichiometry - Absolute number that the chain appear in the Fasta.
                        Sequence - Sequence of the chain

    @Output - Dictionary with the next keys:
              [Name of Chain] - [Name of chain in the interaction] - [Path to that PDB file]
    """

    pairwise_dict = dict()
    pairwise_dict = defaultdict(dict)

    parser = PDBParser(PERMISSIVE=1)
    index = 0

    for pdb_file in glob.glob(os.path.join(TMP_folder,"*.pdb")):
        index += 1
        structure = parser.get_structure('Complex', pdb_file)

        for steichiometry_chain in steichiometry_dict: #A,C
            for chain in structure.get_chains():
                if check_homology(Get_fasta(chain), steichiometry_dict[steichiometry_chain]["sequence"]):
                    chain.real_id = steichiometry_chain

        chain_list = list(structure.get_chains())

        try:
            pairwise_dict[chain_list[0].real_id][chain_list[1].real_id] = (pdb_file)
            pairwise_dict[chain_list[1].real_id][chain_list[0].real_id] = (pdb_file)

        except Exception:
            raise FilesDontMatchException

    return dict(pairwise_dict)

            



import os, sys

class Chain():
    """
    This object stores the properties of a polypeptide chain, 
    such as the name, the aminoacid sequence, the chain of the fasta file which is
    mapped to, and the first and last aminoacid numbers.
    """

    def __init__(self, name, sequence, first_aminoacid, last_aminoacid):
        self.name = name
        self.sequence = str(sequence)
        self.originalChain = list()
        self.first_aminoacid = str(first_aminoacid)
        self.last_aminoacid = str(last_aminoacid)


    def dna_to_placeholder(self):
        """
        This function checks if chain is DNA or Protein, and converts nucleotides to placeholders.
        """
        if all(c in "TGAC" for c in str(self.sequence)):
            new_seq = self.sequence.replace("T","Z")
            new_seq = new_seq.replace("G","B")
            new_seq = new_seq.replace("C","_")
            new_seq = new_seq.replace("A","O")
            new_seq = new_seq.replace("U", "J")
            self.sequence = new_seq
            return True
        else:
            return False

    def pretty_print(self):
        """
        This function prints the information of the Chain object in a more agradable way
        """
        print ("Chain name:", self.name)
        print ("Chain sequence:", self.sequence)
        print ("Chain Original Chain:" , self.originalChain)
        print ("First Aminoacid Number:", self.first_aminoacid)
        print ("Last Aminoacid Number:", self.last_aminoacid)

class Protein_Interaction():
    """
    This class defines an object of a Pairwise Interaction (Protein-Protein, Protein-DNA or DNA-DNA)
    It stores information such as the path of the structure, and if the chains are reversed in order
    according to the fasta.
    """

    def __init__(self, name, path):
        self.name = name
        self.chains =list()
        self.path = path
        self.reversed = False
        self.first_chain = "A"
        self.last_chain = "B"

    def add_chain(self,chain):
        self.chains.append(chain)

    def reverse(self):
        self.first_chain = "B"
        self.last_chain = "A"

    def pretty_print(self):
        print ("Interaction name:", self.name)
        print ("Chains:", self.chains)
        print ("Path:" , self.path)
        print ("Reversed:", self.reversed)


class Query():
    """
    This class is a container for the chains of the query fasta
    """
    def __init__(self, name):
        self.name = name
        self.chains =list()

    def add_chain(self,chain):
        self.chains.append(chain)




import argparse, warnings
from sbiRandM.sbiRandM import *

def mainSuperimp(args):
   if args['verbose']:
      print ("Verbose Mode: Activated.")

   if not os.path.isdir(args['output_folder']):
      if args['verbose']:
         print ('Output folder do not exist. Creating...')
      os.mkdir(args['output_folder'])

   if args['verbose']:
      print ("Checking the steichiometry of the complex using the fasta file.")
   steichiometry_dict = check_fasta_stoichometry(args['fasta_seq'])
   
   if args['verbose']:
      print ("The input fasta consist of", len(steichiometry_dict.keys()), "different chains.")

   if args['verbose']:
      print ('Parsing the pairwise interactions structures from', args['folder'])

   pairwise_dict = obtain_pairwise_dict(steichiometry_dict , args['folder'])
   if args['verbose']:
      print ("Parsed", len(pairwise_dict.keys()), "interaction structures.")

   if args['verbose']:
      print ('Starting to build the complex.')

   try:
      pdb_complex = execute_complex(steichiometry_dict , pairwise_dict , args)
      print ("The program has finished, and your complex is created in " + os.path.join(args['output_folder'],"superimposition_complex.pdb\n\n"))
   except ValueError:
      raise BadFastaException
